@@ -1,6 +1,6 @@
 # Microsoft Azure Linux Agent
 #
-# Copyright 2014 Microsoft Corporation
+# Copyright 2018 Microsoft Corporation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,18 +14,25 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# Requires Python 2.4+ and Openssl 1.0+
+# Requires Python 2.6+ and Openssl 1.0+
 
 import base64
 import json
 import os
 import shutil
 import re
+import sys
+import traceback
 
 import azurelinuxagent.common.conf as conf
+from azurelinuxagent.common.datacontract import get_properties, set_properties, validate_param
+from azurelinuxagent.common.exception import HttpError, ProtocolError
+import azurelinuxagent.common.logger as logger
+from azurelinuxagent.common.utils import restutil
 import azurelinuxagent.common.utils.fileutil as fileutil
 import azurelinuxagent.common.utils.shellutil as shellutil
 import azurelinuxagent.common.utils.textutil as textutil
+from azurelinuxagent.common.telemetryevent import TelemetryEventList
 
 from azurelinuxagent.common.future import httpclient
 from azurelinuxagent.common.protocol.restapi import *
@@ -41,6 +48,8 @@ P7M_FILE_NAME = "Certificates.p7m"
 P7B_FILE_NAME = "Certificates.p7b"
 PEM_FILE_NAME = "Certificates.pem"
 
+IF_NONE_MATCH_HEADER = "If-None-Match"
+
 KEY_AGENT_VERSION_URIS = "versionsManifestUris"
 KEY_URI = "uri"
 
@@ -49,6 +58,14 @@ MAX_PING = 30
 RETRY_PING_INTERVAL = 10
 
 
+def get_traceback(e):
+    if sys.version_info[0] == 3:
+        return e.__traceback__
+    elif sys.version_info[0] == 2:
+        ex_type, ex, tb = sys.exc_info()
+        return tb
+
+
 def _add_content_type(headers):
     if headers is None:
         headers = {}
@@ -81,6 +98,7 @@ class MetadataProtocol(Protocol):
         self.certs = None
         self.agent_manifests = None
         self.agent_etag = None
+        self.cert_etag = None
 
     def _get_data(self, url, headers=None):
         try:
@@ -88,13 +106,21 @@ class MetadataProtocol(Protocol):
         except HttpError as e:
             raise ProtocolError(ustr(e))
 
-        if restutil.request_failed(resp):
+        # NOT_MODIFIED (304) response means the call was successful, so allow that to proceed.
+        is_not_modified = restutil.request_not_modified(resp)
+        if restutil.request_failed(resp) and not is_not_modified:
             raise ProtocolError("{0} - GET: {1}".format(resp.status, url))
 
         data = resp.read()
         etag = resp.getheader('ETag')
+
+        # If the response was 304, then explicilty set data to None
+        if is_not_modified:
+            data = None
+
         if data is not None:
             data = json.loads(ustr(data, encoding="utf-8"))
+
         return data, etag
 
     def _put_data(self, url, data, headers=None):
@@ -123,6 +149,10 @@ class MetadataProtocol(Protocol):
         content = fileutil.read_file(trans_crt_file)
         return textutil.get_bytes_from_pem(content)
 
+    def supports_overprovisioning(self):
+        # Metadata protocol does not support overprovisioning
+        return False
+
     def detect(self):
         self.get_vminfo()
         trans_prv_file = os.path.join(conf.get_lib_dir(),
@@ -140,7 +170,18 @@ class MetadataProtocol(Protocol):
                                 "{0}.crt".format(thumbprint))
         shutil.copyfile(trans_prv_file, prv_file)
         shutil.copyfile(trans_cert_file, crt_file)
-        self.update_goal_state(forced=True)
+        self._update_certs_with_retry()
+
+    def update_host_plugin_from_goal_state(self):
+        # Metadata doesn't cache the goal state; nothing to do here...
+        pass
+
+    def update_goal_state(self):
+        # Metadata doesn't cache the goal state; nothing to do here...
+        pass
+
+    def get_endpoint(self):
+        return self.endpoint
 
     def get_vminfo(self):
         vminfo = VMInfo()
@@ -151,33 +192,44 @@ class MetadataProtocol(Protocol):
     def get_certs(self):
         certlist = CertList()
         certificatedata = CertificateData()
-        data, etag = self._get_data(self.cert_uri)
+        headers = None if self.cert_etag is None else {IF_NONE_MATCH_HEADER: self.cert_etag}
+        data, etag = self._get_data(self.cert_uri, headers=headers)
 
-        set_properties("certlist", certlist, data)
+        if self.cert_etag is None or self.cert_etag != etag:
+            self.cert_etag = etag
 
-        cert_list = get_properties(certlist)
+            set_properties("certlist", certlist, data)
 
-        headers = {
-            "x-ms-vmagent-public-x509-cert": self._get_trans_cert()
-        }
+            cert_list = get_properties(certlist)
 
-        for cert_i in cert_list["certificates"]:
-            certificate_data_uri = cert_i['certificateDataUri']
-            data, etag = self._get_data(certificate_data_uri, headers=headers)
-            set_properties("certificatedata", certificatedata, data)
-            json_certificate_data = get_properties(certificatedata)
+            headers = {
+                "x-ms-vmagent-public-x509-cert": self._get_trans_cert()
+            }
 
-            self.certs = Certificates(self, json_certificate_data)
+            for cert_i in cert_list["certificates"]:
+                certificate_data_uri = cert_i['certificateDataUri']
+                data, etag = self._get_data(certificate_data_uri, headers=headers)
+                set_properties("certificatedata", certificatedata, data)
+                json_certificate_data = get_properties(certificatedata)
+
+                self.certs = Certificates(self, json_certificate_data)
 
         if self.certs is None:
             return None
         return self.certs
 
-    def get_vmagent_manifests(self):
-        self.update_goal_state()
+    def get_incarnation(self):
+        # Always return 0 since Azure Stack does not maintain goal state
+        # incarnation identifiers
+        return 0
 
-        data, etag = self._get_data(self.vmagent_uri)
-        if self.agent_etag is None or self.agent_etag < etag:
+    def get_vmagent_manifests(self):
+        self._update_certs_with_retry()
+
+        headers = None if self.agent_etag is None else {IF_NONE_MATCH_HEADER: self.agent_etag}
+
+        data, etag = self._get_data(self.vmagent_uri, headers=headers)
+        if self.agent_etag is None or self.agent_etag != etag:
             self.agent_etag = etag
 
             # Create a list with a single manifest
@@ -225,13 +277,13 @@ class MetadataProtocol(Protocol):
         return vmagent_pkgs
 
     def get_ext_handlers(self, last_etag=None):
-        self.update_goal_state()
+        self._update_certs_with_retry()
         headers = {
             "x-ms-vmagent-public-x509-cert": self._get_trans_cert()
         }
         ext_list = ExtHandlerList()
         data, etag = self._get_data(self.ext_uri, headers=headers)
-        if last_etag is None or last_etag < etag:
+        if last_etag is None or last_etag != etag:
             set_properties("extensionHandlers", ext_list.extHandlers, data)
         return ext_list, etag
 
@@ -289,16 +341,31 @@ class MetadataProtocol(Protocol):
         certificates = self.get_certs()
         return certificates.cert_list
 
-    def update_goal_state(self, forced=False, max_retry=3):
+    def _update_certs_with_retry(self, max_retry=3):
         # Start updating goalstate, retry on 410
         for retry in range(0, max_retry):
             try:
                 self.update_certs()
                 return
-            except:
+            except Exception as e:
                 logger.verbose("Incarnation is out of date. Update goalstate.")
+                msg = u"Exception updating certs: {0}".format(ustr(e))
+                logger.warn(msg)
+                detailed_msg = '{0} {1}'.format(msg, traceback.extract_tb(get_traceback(e)))
+                logger.verbose(detailed_msg)
         raise ProtocolError("Exceeded max retry updating goal state")
 
+    def download_ext_handler_pkg(self, uri, destination, headers=None, use_proxy=True):
+        success = False
+        try:
+            resp = restutil.http_get(uri, headers=headers, use_proxy=use_proxy)
+            if restutil.request_succeeded(resp):
+                fileutil.write_file(destination, bytearray(resp.read()), asbin=True)
+                success = True
+        except Exception as e:
+            logger.warn("Failed to download from: {0}".format(uri), e)
+        return success
+
 
 class Certificates(object):
     """
