@@ -1,6 +1,6 @@
 # Windows Azure Linux Agent
 #
-# Copyright 2014 Microsoft Corporation
+# Copyright 2018 Microsoft Corporation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,20 +14,23 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# Requires Python 2.4+ and Openssl 1.0+
+# Requires Python 2.6+ and Openssl 1.0+
 #
 
 import glob
 import json
 import os
 import platform
+import random
 import re
 import shutil
 import signal
+import stat
 import subprocess
 import sys
 import time
 import traceback
+import uuid
 import zipfile
 
 from datetime import datetime, timedelta
@@ -37,30 +40,25 @@ import azurelinuxagent.common.logger as logger
 import azurelinuxagent.common.utils.fileutil as fileutil
 import azurelinuxagent.common.utils.restutil as restutil
 import azurelinuxagent.common.utils.textutil as textutil
+from azurelinuxagent.common.cgroupconfigurator import CGroupConfigurator
 
-from azurelinuxagent.common.event import add_event, add_periodic, \
-                                    elapsed_milliseconds, \
-                                    WALAEventOperation
-from azurelinuxagent.common.exception import ProtocolError, \
-                                            ResourceGoneError, \
-                                            UpdateError
+from azurelinuxagent.common.event import add_event, initialize_event_logger_vminfo_common_parameters, elapsed_milliseconds, WALAEventOperation
+from azurelinuxagent.common.exception import ResourceGoneError, UpdateError
 from azurelinuxagent.common.future import ustr
 from azurelinuxagent.common.osutil import get_osutil
-from azurelinuxagent.common.protocol import get_protocol_util
+from azurelinuxagent.common.protocol.util import get_protocol_util
 from azurelinuxagent.common.protocol.hostplugin import HostPluginProtocol
 from azurelinuxagent.common.protocol.wire import WireProtocol
 from azurelinuxagent.common.utils.flexible_version import FlexibleVersion
-from azurelinuxagent.common.version import AGENT_NAME, AGENT_VERSION, AGENT_LONG_VERSION, \
-                                            AGENT_DIR_GLOB, AGENT_PKG_GLOB, \
-                                            AGENT_PATTERN, AGENT_NAME_PATTERN, AGENT_DIR_PATTERN, \
-                                            CURRENT_AGENT, CURRENT_VERSION, \
+from azurelinuxagent.common.version import AGENT_NAME, AGENT_VERSION, AGENT_DIR_PATTERN, \
+                                            CURRENT_AGENT, CURRENT_VERSION, DISTRO_NAME, DISTRO_VERSION, \
                                             is_current_agent_installed
 
 from azurelinuxagent.ga.exthandlers import HandlerManifest
 
 AGENT_ERROR_FILE = "error.json" # File name for agent error record
 AGENT_MANIFEST_FILE = "HandlerManifest.json"
-AGENT_SUPPORTED_FILE = "supported.json"
+AGENT_PARTITION_FILE = "partition"
 
 CHILD_HEALTH_INTERVAL = 15 * 60
 CHILD_LAUNCH_INTERVAL = 5 * 60
@@ -70,10 +68,20 @@ CHILD_POLL_INTERVAL = 60
 MAX_FAILURE = 3 # Max failure allowed for agent before blacklisted
 
 GOAL_STATE_INTERVAL = 3
+GOAL_STATE_INTERVAL_DISABLED = 5 * 60
 
-ORPHAN_WAIT_INTERVAL = 15 * 60 * 60
+ORPHAN_WAIT_INTERVAL = 15 * 60
+
+AGENT_SENTINEL_FILE = "current_version"
+
+READONLY_FILE_GLOBS = [
+    "*.crt",
+    "*.p7m",
+    "*.pem",
+    "*.prv",
+    "ovf-env.xml"
+]
 
-AGENT_SENTINAL_FILE = "current_version"
 
 def get_update_handler():
     return UpdateHandler()
@@ -86,6 +94,8 @@ def get_python_cmd():
 
 class UpdateHandler(object):
 
+    TELEMETRY_HEARTBEAT_PERIOD = timedelta(minutes=30)
+
     def __init__(self):
         self.osutil = get_osutil()
         self.protocol_util = get_protocol_util()
@@ -101,7 +111,10 @@ class UpdateHandler(object):
         self.child_process = None
 
         self.signal_handler = None
-        return
+
+        self.last_telemetry_heartbeat = None
+        self.heartbeat_id = str(uuid.uuid4()).upper()
+        self.heartbeat_counter = 0
 
     def run_latest(self, child_args=None):
         """
@@ -154,20 +167,20 @@ class UpdateHandler(object):
 
             logger.verbose(u"Agent {0} launched with command '{1}'", agent_name, agent_cmd)
 
-            # If the most current agent is the installed agent and update is enabled,
-            # assume updates are likely available and poll every second.
-            # This reduces the start-up impact of finding / launching agent updates on
-            # fresh VMs.
-            if latest_agent is None and conf.get_autoupdate_enabled():
-                poll_interval = 1
-            else:
-                poll_interval = CHILD_POLL_INTERVAL
+            # Setting the poll interval to poll every second to reduce the agent provisioning time;
+            # The daemon shouldn't wait for 60secs before starting the ext-handler in case the
+            # ext-handler kills itself during agent-update during the first 15 mins (CHILD_HEALTH_INTERVAL)
+            poll_interval = 1
 
             ret = None
             start_time = time.time()
             while (time.time() - start_time) < CHILD_HEALTH_INTERVAL:
                 time.sleep(poll_interval)
-                ret = self.child_process.poll()
+                try:
+                    ret = self.child_process.poll()
+                except OSError:
+                    # if child_process has terminated, calling poll could raise an exception
+                    ret = -1
                 if ret is not None:
                     break
 
@@ -181,7 +194,8 @@ class UpdateHandler(object):
                     version=agent_version,
                     op=WALAEventOperation.Enable,
                     is_success=True,
-                    message=msg)
+                    message=msg,
+                    log_event=False)
 
                 if ret is None:
                     ret = self.child_process.wait()
@@ -216,75 +230,139 @@ class UpdateHandler(object):
                     agent_cmd,
                     ustr(e))
                 logger.warn(msg)
+                detailed_message = '{0} {1}'.format(msg, traceback.format_exc())
                 add_event(
                     AGENT_NAME,
                     version=agent_version,
                     op=WALAEventOperation.Enable,
                     is_success=False,
-                    message=msg)
+                    message=detailed_message)
                 if latest_agent is not None:
                     latest_agent.mark_failure(is_fatal=True)
 
         self.child_process = None
         return
 
-    def run(self):
+    def run(self, debug=False):
         """
         This is the main loop which watches for agent and extension updates.
         """
 
-        logger.info(u"Agent {0} is running as the goal state agent",
-                    CURRENT_AGENT)
-
-        # Launch monitoring threads
-        from azurelinuxagent.ga.monitor import get_monitor_handler
-        get_monitor_handler().run()
-
-        from azurelinuxagent.ga.env import get_env_handler
-        get_env_handler().run()
-
-        from azurelinuxagent.ga.exthandlers import get_exthandlers_handler, migrate_handler_state
-        exthandlers_handler = get_exthandlers_handler()
-        migrate_handler_state()
-
         try:
+            logger.info(u"Agent {0} is running as the goal state agent", CURRENT_AGENT)
+
+            #
+            # Fetch the goal state one time; some components depend on information provided by the goal state and this
+            # call ensures the required info is initialized (e.g telemetry depends on the container ID.)
+            #
+            protocol = self.protocol_util.get_protocol()
+            protocol.update_goal_state()
+
+            initialize_event_logger_vminfo_common_parameters(protocol)
+
+            # Log OS-specific info.
+            os_info_msg = u"Distro info: {0} {1}, osutil class being used: {2}, agent service name: {3}"\
+                .format(DISTRO_NAME, DISTRO_VERSION, type(self.osutil).__name__, self.osutil.service_name)
+
+            logger.info(os_info_msg)
+
+            add_event(AGENT_NAME, op=WALAEventOperation.OSInfo, message=os_info_msg)
+
+            # Launch monitoring threads
+            from azurelinuxagent.ga.monitor import get_monitor_handler
+            monitor_thread = get_monitor_handler()
+            monitor_thread.run()
+
+            from azurelinuxagent.ga.env import get_env_handler
+            env_thread = get_env_handler()
+            env_thread.run()
+
+            from azurelinuxagent.ga.exthandlers import get_exthandlers_handler, migrate_handler_state
+            exthandlers_handler = get_exthandlers_handler(protocol)
+            migrate_handler_state()
+
+            from azurelinuxagent.ga.remoteaccess import get_remote_access_handler
+            remote_access_handler = get_remote_access_handler(protocol)
+
             self._ensure_no_orphans()
             self._emit_restart_event()
+            self._ensure_partition_assigned()
+            self._ensure_readonly_files()
+            self._ensure_cgroups_initialized()
+
+            goal_state_interval = GOAL_STATE_INTERVAL if conf.get_extensions_enabled() else GOAL_STATE_INTERVAL_DISABLED
 
             while self.running:
-                if self._is_orphaned:
-                    logger.info("Goal state agent {0} was orphaned -- exiting",
+                #
+                # Check that the parent process (the agent's daemon) is still running
+                #
+                if not debug and self._is_orphaned:
+                    logger.info("Agent {0} is an orphan -- exiting", CURRENT_AGENT)
+                    break
+
+                #
+                # Check that all the threads are still running
+                #
+                if not monitor_thread.is_alive():
+                    logger.warn(u"Monitor thread died, restarting")
+                    monitor_thread.start()
+
+                if not env_thread.is_alive():
+                    logger.warn(u"Environment thread died, restarting")
+                    env_thread.start()
+
+                #
+                # Process the goal state
+                #
+                goal_state_fetched = False
+                try:
+                    protocol.update_goal_state()
+                    goal_state_fetched = True
+                except Exception as e:
+                    msg = u"Exception retrieving the goal state: {0}".format(ustr(traceback.format_exc()))
+                    add_event(AGENT_NAME, op=WALAEventOperation.FetchGoalState, version=CURRENT_VERSION, is_success=False, message=msg)
+
+                if goal_state_fetched:
+                    if self._upgrade_available(protocol):
+                        available_agent = self.get_latest_agent()
+                        if available_agent is None:
+                            logger.info(
+                                "Agent {0} is reverting to the installed agent -- exiting",
                                 CURRENT_AGENT)
-                    break
+                        else:
+                            logger.info(
+                                u"Agent {0} discovered update {1} -- exiting",
+                                CURRENT_AGENT,
+                                available_agent.name)
+                        break
 
-                if self._upgrade_available():
-                    if len(self.agents) > 0:
-                        logger.info(
-                            u"Agent {0} discovered {1} as an update and will exit",
-                            CURRENT_AGENT,
-                            self.agents[0].name)
-                    break
+                    utc_start = datetime.utcnow()
 
-                utc_start = datetime.utcnow()
+                    last_etag = exthandlers_handler.last_etag
+                    exthandlers_handler.run()
 
-                last_etag = exthandlers_handler.last_etag
-                exthandlers_handler.run()
+                    remote_access_handler.run()
 
-                if last_etag != exthandlers_handler.last_etag:
-                    add_event(
-                        AGENT_NAME,
-                        version=CURRENT_VERSION,
-                        op=WALAEventOperation.ProcessGoalState,
-                        is_success=True,
-                        duration=elapsed_milliseconds(utc_start),
-                        log_event=True)
+                    if last_etag != exthandlers_handler.last_etag:
+                        self._ensure_readonly_files()
+                        duration = elapsed_milliseconds(utc_start)
+                        logger.info('ProcessGoalState completed [incarnation {0}; {1} ms]',
+                                    exthandlers_handler.last_etag,
+                                    duration)
+                        add_event(
+                            AGENT_NAME,
+                            op=WALAEventOperation.ProcessGoalState,
+                            duration=duration,
+                            message="Incarnation {0}".format(exthandlers_handler.last_etag))
 
-                time.sleep(GOAL_STATE_INTERVAL)
+                self._send_heartbeat_telemetry(protocol)
+
+                time.sleep(goal_state_interval)
 
         except Exception as e:
-            logger.warn(u"Agent {0} failed with exception: {1}",
-                        CURRENT_AGENT,
-                        ustr(e))
+            msg = u"Agent {0} failed with exception: {1}".format(CURRENT_AGENT, ustr(e))
+            self._set_sentinel(msg=msg)
+            logger.warn(msg)
             logger.warn(traceback.format_exc())
             sys.exit(1)
             # additional return here because sys.exit is mocked in unit tests
@@ -294,12 +372,7 @@ class UpdateHandler(object):
         sys.exit(0)
 
     def forward_signal(self, signum, frame):
-        # Note:
-        #  - At present, the handler is registered only for SIGTERM.
-        #    However, clean shutdown is both SIGTERM and SIGKILL.
-        #    A SIGKILL handler is not being registered at this time to
-        #    minimize perturbing the code.
-        if signum in (signal.SIGTERM, signal.SIGKILL):
+        if signum == signal.SIGTERM:
             self._shutdown()
 
         if self.child_process is None:
@@ -310,13 +383,14 @@ class UpdateHandler(object):
             CURRENT_AGENT,
             signum,
             self.child_agent.name if self.child_agent is not None else CURRENT_AGENT)
+
         self.child_process.send_signal(signum)
 
         if self.signal_handler not in (None, signal.SIG_IGN, signal.SIG_DFL):
             self.signal_handler(signum, frame)
         elif self.signal_handler is signal.SIG_DFL:
             if signum == signal.SIGTERM:
-                # TODO: This should set self.running to False vs. just exiting
+                self._shutdown()
                 sys.exit(0)
         return
 
@@ -338,17 +412,20 @@ class UpdateHandler(object):
         return available_agents[0] if len(available_agents) >= 1 else None
 
     def _emit_restart_event(self):
-        if not self._is_clean_start:
-            msg = u"{0} did not terminate cleanly".format(CURRENT_AGENT)
-            logger.info(msg)
-            add_event(
-                AGENT_NAME,
-                version=CURRENT_VERSION,
-                op=WALAEventOperation.Restart,
-                is_success=False,
-                message=msg)
+        try:
+            if not self._is_clean_start:
+                msg = u"Agent did not terminate cleanly: {0}".format(
+                            fileutil.read_file(self._sentinel_file_path()))
+                logger.info(msg)
+                add_event(
+                    AGENT_NAME,
+                    version=CURRENT_VERSION,
+                    op=WALAEventOperation.Restart,
+                    is_success=False,
+                    message=msg)
+        except Exception:
+            pass
 
-        self._set_sentinal() 
         return
 
     def _ensure_no_orphans(self, orphan_wait_interval=ORPHAN_WAIT_INTERVAL):
@@ -382,6 +459,32 @@ class UpdateHandler(object):
                     ustr(e))
         return
 
+    def _ensure_partition_assigned(self):
+        """
+        Assign the VM to a partition (0 - 99). Downloaded updates may be configured
+        to run on only some VMs; the assigned partition determines eligibility.
+        """
+        if not os.path.exists(self._partition_file):
+            partition = ustr(int(datetime.utcnow().microsecond / 10000))
+            fileutil.write_file(self._partition_file, partition)
+            add_event(
+                AGENT_NAME,
+                version=CURRENT_VERSION,
+                op=WALAEventOperation.Partition,
+                is_success=True,
+                message=partition)
+
+    def _ensure_readonly_files(self):
+        for g in READONLY_FILE_GLOBS:
+            for path in glob.iglob(os.path.join(conf.get_lib_dir(), g)):
+                os.chmod(path, stat.S_IRUSR)
+
+    def _ensure_cgroups_initialized(self):
+        configurator = CGroupConfigurator.get_instance()
+        configurator.create_agent_cgroups(track_cgroups=True)
+        configurator.cleanup_legacy_cgroups()
+        configurator.create_extension_cgroups_root()
+
     def _evaluate_agent_health(self, latest_agent):
         """
         Evaluate the health of the selected agent: If it is restarting
@@ -412,7 +515,6 @@ class UpdateHandler(object):
 
     def _filter_blacklisted_agents(self):
         self.agents = [agent for agent in self.agents if not agent.is_blacklisted]
-        return
 
     def _find_agents(self):
         """
@@ -427,10 +529,7 @@ class UpdateHandler(object):
 
     def _get_host_plugin(self, protocol=None):
         return protocol.client.get_host_plugin() \
-                                if protocol and \
-                                    type(protocol) is WireProtocol and \
-                                    protocol.client \
-                                else None
+            if protocol and type(protocol) is WireProtocol and protocol.client else None
 
     def _get_pid_parts(self):
         pid_file = conf.get_agent_pid_file_path()
@@ -447,19 +546,7 @@ class UpdateHandler(object):
 
     @property
     def _is_clean_start(self):
-        if not os.path.isfile(self._sentinal_file_path()):
-            return True
-
-        try:
-            if fileutil.read_file(self._sentinal_file_path()) != CURRENT_AGENT:
-                return True
-        except Exception as e:
-            logger.warn(
-                u"Exception reading sentinal file {0}: {1}",
-                self._sentinal_file_path(),
-                str(e))
-
-        return False
+        return not os.path.isfile(self._sentinel_file_path())
 
     @property
     def _is_orphaned(self):
@@ -472,11 +559,29 @@ class UpdateHandler(object):
 
         return fileutil.read_file(conf.get_agent_pid_file_path()) != ustr(parent_pid)
 
+    def _is_version_eligible(self, version):
+        # Ensure the installed version is always eligible
+        if version == CURRENT_VERSION and is_current_agent_installed():
+            return True
+
+        for agent in self.agents:
+            if agent.version == version:
+                return agent.is_available
+
+        return False
+
     def _load_agents(self):
         path = os.path.join(conf.get_lib_dir(), "{0}-*".format(AGENT_NAME))
         return [GuestAgent(path=agent_dir)
                         for agent_dir in glob.iglob(path) if os.path.isdir(agent_dir)]
 
+    def _partition(self):
+        return int(fileutil.read_file(self._partition_file))
+
+    @property
+    def _partition_file(self):
+        return os.path.join(conf.get_lib_dir(), AGENT_PARTITION_FILE)
+
     def _purge_agents(self):
         """
         Remove from disk all directories and .zip files of unknown agents
@@ -485,8 +590,8 @@ class UpdateHandler(object):
         path = os.path.join(conf.get_lib_dir(), "{0}-*".format(AGENT_NAME))
 
         known_versions = [agent.version for agent in self.agents]
-        if not is_current_agent_installed() and CURRENT_VERSION not in known_versions:
-            logger.warn(
+        if CURRENT_VERSION not in known_versions:
+            logger.verbose(
                 u"Running Agent {0} was not found in the agent manifest - adding to list",
                 CURRENT_VERSION)
             known_versions.append(CURRENT_VERSION)
@@ -511,35 +616,37 @@ class UpdateHandler(object):
         self.agents.sort(key=lambda agent: agent.version, reverse=True)
         return
 
-    def _set_sentinal(self, agent=CURRENT_AGENT):
+    def _set_sentinel(self, agent=CURRENT_AGENT, msg="Unknown cause"):
         try:
-            fileutil.write_file(self._sentinal_file_path(), agent)
+            fileutil.write_file(
+                self._sentinel_file_path(),
+                "[{0}] [{1}]".format(agent, msg))
         except Exception as e:
             logger.warn(
-                u"Exception writing sentinal file {0}: {1}",
-                self._sentinal_file_path(),
+                u"Exception writing sentinel file {0}: {1}",
+                self._sentinel_file_path(),
                 str(e))
         return
 
-    def _sentinal_file_path(self):
-        return os.path.join(conf.get_lib_dir(), AGENT_SENTINAL_FILE)
+    def _sentinel_file_path(self):
+        return os.path.join(conf.get_lib_dir(), AGENT_SENTINEL_FILE)
 
     def _shutdown(self):
         self.running = False
 
-        if not os.path.isfile(self._sentinal_file_path()):
+        if not os.path.isfile(self._sentinel_file_path()):
             return
 
         try:
-            os.remove(self._sentinal_file_path())
+            os.remove(self._sentinel_file_path())
         except Exception as e:
             logger.warn(
-                u"Exception removing sentinal file {0}: {1}",
-                self._sentinal_file_path(),
+                u"Exception removing sentinel file {0}: {1}",
+                self._sentinel_file_path(),
                 str(e))
         return
 
-    def _upgrade_available(self, base_version=CURRENT_VERSION):
+    def _upgrade_available(self, protocol, base_version=CURRENT_VERSION):
         # Emit an event expressing the state of AutoUpdate
         # Note:
         # - Duplicate events get suppressed; state transitions always emit
@@ -547,10 +654,12 @@ class UpdateHandler(object):
             AGENT_NAME,
             version=CURRENT_VERSION,
             op=WALAEventOperation.AutoUpdate,
-            is_success=conf.get_autoupdate_enabled())
+            is_success=conf.get_autoupdate_enabled(),
+            log_event=False)
 
         # Ignore new agents if updating is disabled
         if not conf.get_autoupdate_enabled():
+            logger.warn(u"Agent auto-update is disabled.")
             return False
 
         now = time.time()
@@ -563,61 +672,45 @@ class UpdateHandler(object):
             return False
 
         family = conf.get_autoupdate_gafamily()
-        logger.verbose("Checking for agent family {0} updates", family)
+        logger.info("Checking for agent updates (family: {0})", family)
 
         self.last_attempt_time = now
-        protocol = self.protocol_util.get_protocol()
 
-        for update_goal_state in [False, True]:
-            try:
-                if update_goal_state:
-                    protocol.update_goal_state(forced=True)
+        try:
+            manifest_list, etag = protocol.get_vmagent_manifests()
 
-                manifest_list, etag = protocol.get_vmagent_manifests()
-
-                manifests = [m for m in manifest_list.vmAgentManifests \
-                                if m.family == family and \
-                                    len(m.versionsManifestUris) > 0]
-                if len(manifests) == 0:
-                    logger.verbose(u"Incarnation {0} has no {1} agent updates",
-                                    etag, family)
-                    return False
-
-                pkg_list = protocol.get_vmagent_pkgs(manifests[0])
-
-                # Set the agents to those available for download at least as
-                # current as the existing agent and remove from disk any agent
-                # no longer reported to the VM.
-                # Note:
-                #  The code leaves on disk available, but blacklisted, agents
-                #  so as to preserve the state. Otherwise, those agents could be
-                #  again downloaded and inappropriately retried.
-                host = self._get_host_plugin(protocol=protocol)
-                self._set_agents([GuestAgent(pkg=pkg, host=host) \
-                                     for pkg in pkg_list.versions])
-
-                self._purge_agents()
-                self._filter_blacklisted_agents()
-
-                # Return True if more recent agents are available
-                return len(self.agents) > 0 and \
-                            self.agents[0].version > base_version
-
-            except Exception as e:
-                if isinstance(e, ResourceGoneError):
-                    continue
-
-                msg = u"Exception retrieving agent manifests: {0}".format(
-                            ustr(e))
-                logger.warn(msg)
-                add_event(
-                    AGENT_NAME,
-                    op=WALAEventOperation.Download,
-                    version=CURRENT_VERSION,
-                    is_success=False,
-                    message=msg)
+            manifests = [m for m in manifest_list.vmAgentManifests \
+                            if m.family == family and len(m.versionsManifestUris) > 0]
+            if len(manifests) == 0:
+                logger.verbose(u"Incarnation {0} has no {1} agent updates",
+                                etag, family)
                 return False
 
+            pkg_list = protocol.get_vmagent_pkgs(manifests[0])
+
+            # Set the agents to those available for download at least as
+            # current as the existing agent and remove from disk any agent
+            # no longer reported to the VM.
+            # Note:
+            #  The code leaves on disk available, but blacklisted, agents
+            #  so as to preserve the state. Otherwise, those agents could be
+            #  again downloaded and inappropriately retried.
+            host = self._get_host_plugin(protocol=protocol)
+            self._set_agents([GuestAgent(pkg=pkg, host=host) for pkg in pkg_list.versions])
+
+            self._purge_agents()
+            self._filter_blacklisted_agents()
+
+            # Return True if current agent is no longer available or an
+            # agent with a higher version number is available
+            return not self._is_version_eligible(base_version) \
+                or (len(self.agents) > 0 and self.agents[0].version > base_version)
+
+        except Exception as e:
+            msg = u"Exception retrieving agent manifests: {0}".format(ustr(traceback.format_exc()))
+            add_event(AGENT_NAME, op=WALAEventOperation.Download, version=CURRENT_VERSION, is_success=False, message=msg)
+            return False
+
     def _write_pid_file(self):
         pid_files = self._get_pid_files()
 
@@ -644,6 +737,21 @@ class UpdateHandler(object):
 
         return pid_files, pid_file
 
+    def _send_heartbeat_telemetry(self, protocol):
+        if self.last_telemetry_heartbeat is None:
+            self.last_telemetry_heartbeat = datetime.utcnow() - UpdateHandler.TELEMETRY_HEARTBEAT_PERIOD
+
+        if datetime.utcnow() >= (self.last_telemetry_heartbeat + UpdateHandler.TELEMETRY_HEARTBEAT_PERIOD):
+            dropped_packets = self.osutil.get_firewall_dropped_packets(protocol.get_endpoint())
+            msg = "{0};{1};{2}".format(self.heartbeat_counter, self.heartbeat_id, dropped_packets)
+
+            add_event(name=AGENT_NAME, version=CURRENT_VERSION, op=WALAEventOperation.HeartBeat, is_success=True,
+                      message=msg, log_event=False)
+            self.heartbeat_counter += 1
+
+            logger.info(u"[HEARTBEAT] Agent {0} is running as the goal state agent", CURRENT_AGENT)
+            self.last_telemetry_heartbeat = datetime.utcnow()
+
 
 class GuestAgent(object):
     def __init__(self, path=None, pkg=None, host=None):
@@ -667,8 +775,6 @@ class GuestAgent(object):
 
         self.error = GuestAgentError(self.get_agent_error_file())
         self.error.load()
-        self.supported = Supported(self.get_agent_supported_file())
-        self.supported.load()
 
         try:
             self._ensure_downloaded()
@@ -677,6 +783,13 @@ class GuestAgent(object):
             if isinstance(e, ResourceGoneError):
                 raise
 
+            # The agent was improperly blacklisting versions due to a timeout
+            # encountered while downloading a later version. Errors of type
+            # socket.error are IOError, so this should provide sufficient
+            # protection against a large class of I/O operation failures.
+            if isinstance(e, IOError):
+                raise
+
             # Note the failure, blacklist the agent if the package downloaded
             # - An exception with a downloaded package indicates the package
             #   is corrupt (e.g., missing the HandlerManifest.json file)
@@ -691,7 +804,6 @@ class GuestAgent(object):
                 op=WALAEventOperation.Install,
                 is_success=False,
                 message=msg)
-        return
 
     @property
     def name(self):
@@ -712,13 +824,9 @@ class GuestAgent(object):
     def get_agent_pkg_path(self):
         return ".".join((os.path.join(conf.get_lib_dir(), self.name), "zip"))
 
-    def get_agent_supported_file(self):
-        return os.path.join(conf.get_lib_dir(), self.name, AGENT_SUPPORTED_FILE)
-
     def clear_error(self):
         self.error.clear()
         self.error.save()
-        return
 
     @property
     def is_available(self):
@@ -730,15 +838,8 @@ class GuestAgent(object):
 
     @property
     def is_downloaded(self):
-        return self.is_blacklisted or os.path.isfile(self.get_agent_manifest_path())
-
-    @property
-    def _is_optional(self):
-        return self.error is not None and self.error.is_sentinel and self.supported.is_supported
-
-    @property
-    def _in_slice(self):
-        return self.supported.is_supported and self.supported.in_slice
+        return self.is_blacklisted or \
+                os.path.isfile(self.get_agent_manifest_path())
 
     def mark_failure(self, is_fatal=False):
         try:
@@ -750,21 +851,6 @@ class GuestAgent(object):
                 logger.warn(u"Agent {0} is permanently blacklisted", self.name)
         except Exception as e:
             logger.warn(u"Agent {0} failed recording error state: {1}", self.name, ustr(e))
-        return
-
-    def _enable(self):
-        # Enable optional agents if within the "slice"
-        # - The "slice" is a percentage of the agent to execute
-        # - Blacklist out-of-slice agents to prevent reconsideration
-        if self._is_optional:
-            if self._in_slice:
-                self.error.clear()
-                self.error.save()
-                logger.info(u"Enabled optional Agent {0}", self.name)
-            else:
-                self.mark_failure(is_fatal=True)
-                logger.info(u"Optional Agent {0} not in slice", self.name)
-        return
 
     def _ensure_downloaded(self):
         logger.verbose(u"Ensuring Agent {0} is downloaded", self.name)
@@ -788,18 +874,15 @@ class GuestAgent(object):
             op=WALAEventOperation.Install,
             is_success=True,
             message=msg)
-        return
 
     def _ensure_loaded(self):
         self._load_manifest()
         self._load_error()
-        self._load_supported()
-
-        self._enable()
-        return
 
     def _download(self):
-        for uri in self.pkg.uris:
+        uris_shuffled = self.pkg.uris
+        random.shuffle(uris_shuffled)
+        for uri in uris_shuffled:
             if not HostPluginProtocol.is_default_channel() and self._fetch(uri.uri):
                 break
 
@@ -838,11 +921,11 @@ class GuestAgent(object):
                 message=msg)
             raise UpdateError(msg)
 
-        return
-
     def _fetch(self, uri, headers=None, use_proxy=True):
         package = None
         try:
+            is_healthy = True
+            error_response = ''
             resp = restutil.http_get(uri, use_proxy=use_proxy, headers=headers)
             if restutil.request_succeeded(resp):
                 package = resp.read()
@@ -851,8 +934,13 @@ class GuestAgent(object):
                                     asbin=True)
                 logger.verbose(u"Agent {0} downloaded from {1}", self.name, uri)
             else:
-                logger.verbose("Fetch was unsuccessful [{0}]",
-                               restutil.read_response_error(resp))
+                error_response = restutil.read_response_error(resp)
+                logger.verbose("Fetch was unsuccessful [{0}]", error_response)
+                is_healthy = not restutil.request_failed_at_hostplugin(resp)
+
+            if self.host is not None:
+                self.host.report_fetch_health(uri, is_healthy, source='GuestAgent', response=error_response)
+
         except restutil.HttpError as http_error:
             if isinstance(http_error, ResourceGoneError):
                 raise
@@ -871,14 +959,6 @@ class GuestAgent(object):
             logger.verbose(u"Agent {0} error state: {1}", self.name, ustr(self.error))
         except Exception as e:
             logger.warn(u"Agent {0} failed loading error state: {1}", self.name, ustr(e))
-        return
-
-    def _load_supported(self):
-        try:
-            self.supported = Supported(self.get_agent_supported_file())
-            self.supported.load()
-        except Exception as e:
-            self.supported = Supported()
 
     def _load_manifest(self):
         path = self.get_agent_manifest_path()
@@ -976,10 +1056,6 @@ class GuestAgentError(object):
     def is_blacklisted(self):
         return self.was_fatal or self.failure_count >= MAX_FAILURE
 
-    @property
-    def is_sentinel(self):
-        return self.was_fatal and self.last_failure == 0.0
-
     def load(self):
         if self.path is not None and os.path.isfile(self.path):
             with open(self.path, 'r') as f:
@@ -1015,60 +1091,3 @@ class GuestAgentError(object):
             self.last_failure,
             self.failure_count,
             self.was_fatal)
-
-class Supported(object):
-    def __init__(self, path):
-        if path is None:
-            raise UpdateError(u"Supported requires a path")
-        self.path = path
-        self.distributions = {}
-        return
-
-    @property
-    def is_supported(self):
-        return self._supported_distribution is not None
-
-    @property
-    def in_slice(self):
-        d = self._supported_distribution
-        return d is not None and d.in_slice
-
-    def load(self):
-        self.distributions = {}
-        try:
-            if self.path is not None and os.path.isfile(self.path):
-                j = json.loads(fileutil.read_file(self.path))
-                for d in j:
-                    self.distributions[d] = SupportedDistribution(j[d])
-        except Exception as e:
-            logger.warn("Failed JSON parse of {0}: {1}".format(self.path, e))
-        return
-
-    @property
-    def _supported_distribution(self):
-        for d in self.distributions:
-            dd = self.distributions[d]
-            if dd.is_supported:
-                return dd
-        return None
-
-class SupportedDistribution(object):
-    def __init__(self, s):
-        if s is None or not isinstance(s, dict):
-            raise UpdateError(u"SupportedDisribution requires a dictionary")
-
-        self.slice = s['slice']
-        self.versions = s['versions']
-
-    @property
-    def is_supported(self):
-        d = ','.join(platform.linux_distribution())
-        for v in self.versions:
-            if re.match(v, d):
-                return True
-        return False
-
-    @property
-    def in_slice(self):
-        n = int((60 * self.slice) / 100)
-        return (n - datetime.utcnow().second) > 0
