@@ -16,7 +16,8 @@
 #
 from azurelinuxagent.common.osutil.suse import SUSE11OSUtil
 from .test_default import osutil_get_dhcp_pid_should_return_a_list_of_pids
-from tests.tools import *
+from tests.tools import AgentTestCase
+import unittest
 
 
 class TestSUSE11OSUtil(AgentTestCase):
